from optparse import OptionParser
import os
import sys

import ASTPrinter
import File
import fslparser
import lexer
import HookHandler
import Interpreter
import utils

class PrintHookHandler(HookHandler.HookHandler):
    def __init__(self, format_absolute, format_quote, print_details, print_stats,
            out_filename=None, stats_limit=1, printing_relative_dir=None):

        self.format_absolute = format_absolute
        self.format_quote = format_quote
        self.print_details = print_details
        self.print_stats = print_stats
        self.out_filename = out_filename
        self.stats_limit = stats_limit

        if printing_relative_dir is None:
            self.printing_relative_dir = os.getcwd()
        else:
            self.printing_relative_dir = printing_relative_dir

        # We print basic file list if
        # - print_details or print_stats is not set,
        #   i.e. program was run with basic options    OR
        # - out_filename is set
        self.print_basic = (not (print_details or print_stats)) or out_filename

        if self.out_filename:
            self.outfile = file(out_filename, 'wt')
        else:
            self.outfile = sys.stdout

        # Stuff for print_stats.
        self._file_ext_dict = {}     # {file extension: file count, byte count}
        self._directory_dict = {}    # {directory: file count, byte count)
        self._fileobjs = []      # Included files
        self._total_files = 0
        self._total_bytes = 0

    def end(self):
        if self.out_filename:
            self.outfile.close()
        if self.print_stats:
            self._do_print_stats()

    def _do_print_stats(self):
        if self._total_files == 0:
            avgsize = 0
        else:
            avgsize = self._total_bytes / float(self._total_files)
        print
        print 'TOTAL   : %5d files, %8s, avg size %8s' % (self._total_files,
            utils.format_size(self._total_bytes), utils.format_size(avgsize))
        print

        extlist = [(bytes, files, ext) for ext, (files, bytes) in self._file_ext_dict.items()]
        extlist.sort()
        extlist.reverse()
        for bytes, files, ext in extlist[:10]:
            if files > 0:
                avgsize = bytes / float(files)
                percent = bytes / float(self._total_bytes) * 100
                print '*.%-6s: %5d files, %8s, avg size %8s, %2.0f%%' % (ext, files,
                    utils.format_size(bytes), utils.format_size(avgsize), percent)

        dirlist = [(bytes, files, dirname) for dirname, (files, bytes) in self._directory_dict.items()]
        dirlist.sort()
        dirlist.reverse()
        print
        print '* Directories *'
        for bytes, files, dirname in dirlist[:10]:
            if files > 0:
                avgsize = bytes / float(files)
                percent = bytes / float(self._total_bytes) * 100
                format_dirname = self._format_filename(File.File(dirname))
                print '%3d files, %8s, %2.0f%% %s' % (files, utils.format_size(bytes), percent, format_dirname)

        filelist = [(fileobj.get_size(), fileobj) for fileobj in self._fileobjs]
        filelist.sort()
        filelist.reverse()
        print
        print '* Files *'
        for bytes, fileobj in filelist[:10]:
            print '%-8s %s' % (utils.format_size(bytes), self._format_filename(fileobj))

    def _format_filename(self, fileobj):
        if self.format_absolute:
            filename = os.path.abspath(fileobj.abs_filename)
        else:
            filename = fileobj.calc_relative_filename(self.printing_relative_dir)
            if filename is None:
                msg = "Can't print file '%s' as relative filename" % fileobj.abs_filename
                raise Interpreter.FSLRuntimeError(msg)
        if self.format_quote:
            filename = '"%s"'% filename
        return filename

    def include_file(self, fileobj, rule):
        if self.print_details:
            print '+ %s' % self._format_filename(fileobj)
            print '  %s bytes, %s' % (fileobj.get_size(), fileobj.get_modtime())
        if self.print_basic:
            print >>self.outfile, self._format_filename(fileobj)
        if self.print_stats:
            self._total_files += 1
            self._total_bytes += fileobj.get_size()
            self._fileobjs.append(fileobj)

            root, extension = os.path.splitext(fileobj.abs_filename)
            extension = extension.lower()
            if extension.startswith('.'):
                extension = extension[1:]
            ext_files, ext_bytes = self._file_ext_dict.get(extension, (0, 0))
            self._file_ext_dict[extension] = ext_files+1, ext_bytes+fileobj.get_size()

            dirname, tail = os.path.split(fileobj.abs_filename)
            dir_files, dir_bytes = self._directory_dict.get(dirname, (0, 0))
            self._directory_dict[dirname] = dir_files+1, dir_bytes+fileobj.get_size()

    def exclude_file(self, fileobj, rule):
        if self.print_details:
            print '- %s' % self._format_filename(fileobj)

    def start_scan(self, dirnameobj, glob_pattern):
        if self.print_details:
            print 'D %s (glob "%s")' % (self._format_filename(dirnameobj), glob_pattern)

    def end_scan(self, dirnameobj, files_count, bytes_count):
        if self.print_details:
            print 'end(D) %s (%s files, %s)' % (self._format_filename(dirnameobj), files_count, utils.format_size(bytes_count))
        elif self.print_stats and (bytes_count >= self.stats_limit):
            print '%4d files, %8s %s' % (files_count, utils.format_size(bytes_count), self._format_filename(dirnameobj))
#        print self.print_stats, self.stats_limit, bytes_count

def main(argv):
    argv = filter(lambda a: a, argv) # Remove empty strings

    USAGE = 'fsltool.py [options] <source-filenames>'
    parser = OptionParser(usage=USAGE)
#    parser.add_option('-a', action='store_true', dest='format_absolute',
#        help="Print absolute file names")
    parser.add_option('--ast', action='store_true', dest='print_ast',
        help="Print Abstract Syntax Tree only, don't run interpreter")
    parser.add_option('-d', action='store_true', dest='details',
        help='Print detailed information')
    parser.add_option('-l', action='store_true', dest='format_relative',
        help='Print reLative filename instead of absolute')
    parser.add_option('--lexer', action='store_true', dest='print_lexer',
        help="Print lexer output only, don't run interpreter")
    parser.add_option('-o', type='string', dest='outfile', default=None,
        help='Write file list into file instead of stdout')
    parser.add_option('-q', action='store_true', dest='format_quote',
        help='Surround file name with quote marks')
    parser.add_option('-r', type='string', dest='rootdir', default='.',
        help='Root directory for relative filenames [.]')
    parser.add_option('--reldir', type='string', dest='printing_relative_dir', default=None,
        help='With -l, print filenames relative to this directory instead of "."')
    parser.add_option('-s', action='store_true', dest='stats',
        help='Print statistics about the files included')
    parser.add_option('-t', type='int', dest='stats_limit', default=1,
        help='Threshold in bytes for -s to print directory stats at scan time [1]')

    options, args = parser.parse_args(argv)

    if not args:
        parser.print_usage()
        return
    filenames = args

    if options.print_lexer:
        for filename in filenames:
            for token in lexer.iter_tokens(filename):
                print 'Line %2d, type %-11s "%s"' % (
                    token.lineno, lexer.type2str(token.tokentype), token.strvalue)
        return

    ast = fslparser.parse(filenames, options.rootdir)

    if options.print_ast:
        ast.visit(ASTPrinter.ASTPrinter())
        return

    if not os.path.exists(options.rootdir):
        print 'Root dir "%s" not found' % options.rootdir
        return
    if not os.path.isdir(options.rootdir):
        print 'Root dir "%s" is not a directory' % options.rootdir
        return

    printing_relative_dir = options.printing_relative_dir
    if printing_relative_dir is None:
        printing_relative_dir = os.getcwd()
    printing_relative_dir = utils.abspath(printing_relative_dir)

    if options.format_relative:
        common_prefix = ast.innerdir_common_prefix()
        if not common_prefix.startswith(printing_relative_dir):
            print "Can't print relative filenames: '%s' is not a common prefix for the files" % printing_relative_dir
            return

    hook_handler = PrintHookHandler(not options.format_relative, options.format_quote, options.details,
        options.stats, options.outfile, options.stats_limit, printing_relative_dir)

    interpreter = Interpreter.Interpreter(ast)
    interpreter.add_hook_handler(hook_handler)
    interpreter.run()

    hook_handler.end()

if __name__=='__main__':
    import sys
    main(sys.argv[1:])