import cStringIO
import os
import sys
import unittest

import fsl.fsltool, fsl.utils
import test_interpreter
import tutils

# See notes in test_interpreter.py on the following lines.
try:
    set, frozenset
except NameError:
    import sets
    set = sets.Set
    frozenset = sets.ImmutableSet

OUT_FILENAME = 'tmpfilelist'

class TestFsltool(unittest.TestCase):

    def setUp(self):
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        self.stdout = cStringIO.StringIO()
        self.stderr = cStringIO.StringIO()
        sys.stdout = self.stdout
        sys.stderr = self.stderr

    def tearDown(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    def _call_fsltool(self, args):
        fsl.fsltool.main(args)
        return self.stdout.getvalue()

    def _check_filenames(self, actual_filenames, expected_filenames, expected_rootdir, source_msg, expect_abspath=True):
        for filename in actual_filenames:
            self.failUnless(os.path.exists(filename),
                'Interpreter produced "%s", but the file was not found by os.path.exists' % filename)

        actual_set = set(actual_filenames)

        expected_set = set()
        for exp_filename in expected_filenames:
            realexp = fsl.utils.combine_rootdir(expected_rootdir, exp_filename)
            if expect_abspath:
                realexp = os.path.abspath(realexp)
            expected_set.add(realexp)

        if expected_set != actual_set:
            missing_set = expected_set - actual_set
            extraneous_set = actual_set - expected_set

            msg = '"%s" didn\'t produce expected files' % source_msg

            if missing_set:
                missing_list = list(missing_set)
                missing_list.sort()
                missing_str = ', '.join(['"%s"' % s for s in missing_list])
                msg += '\nMissing: ' + missing_str

            if extraneous_set:
                extraneous_list = list(extraneous_set)
                extraneous_list.sort()
                extraneous_str = ', '.join(['"%s"' % s for s in extraneous_list])
                msg += '\nExtraneous: ' + extraneous_str

            self.fail(msg)

    def _check_enough_output(self, expbytes, output, message):
        self.failIf(len(output) < expbytes,
            '%s produced too little output (expected %d bytes, got %d).\nOutput="%s"' % (
                message, expbytes, len(output), output))

    def test_noargs(self):
        """Test that fsltool prints usage message when
        called with no arguments."""
        output = self._call_fsltool([])
        self.failUnless('usage' in output.lower(),
            "fsltool didn't print the word 'usage' when called with no args\nOutput: '%s'" % output)

    def test_allfiles_plain(self):
        """Test that fsltool prints all filenames when given a
        include-all ruleset."""
        rulefile = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-r', tutils.ROOTDIR, rulefile])
        filenames = output.splitlines()
        self._check_filenames(filenames, test_interpreter.ALL_FILES, tutils.ROOTDIR, rulefile)

    def test_allfiles_quoted(self):
        """Test that fsltool -q prints all filenames when given a
        include-all ruleset."""

        rulefile = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-q', '-r', tutils.ROOTDIR, rulefile])
        filenames = []
        for filename in output.splitlines():
            if len(filename) < 2:
                filenames.append(filename)
            else:
                filenames.append(filename[1:-1])
        self._check_filenames(filenames, test_interpreter.ALL_FILES, tutils.ROOTDIR, rulefile)

    def test_allfiles_relative(self):
        """Test that fsltool -l prints all filenames when given a
        include-all ruleset."""
        rulefile = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-l', '-r', tutils.ROOTDIR, rulefile])
        filenames = output.splitlines()
        self._check_filenames(filenames, test_interpreter.ALL_FILES, tutils.ROOTDIR, rulefile, False)

    def test_allfiles_outfile(self):
        """Test that fsltool -o writes filenames to output file."""

        if os.path.exists(OUT_FILENAME):
            os.remove(OUT_FILENAME)

        rulefile = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-o', OUT_FILENAME, '-r', tutils.ROOTDIR, rulefile])

        self.failUnless(os.path.exists(OUT_FILENAME),
            "fsltool -o %s didn't produce output file" % OUT_FILENAME)

        infile = file(OUT_FILENAME, 'rt')
        filenames = infile.read().splitlines()
        infile.close()

        self._check_filenames(filenames, test_interpreter.ALL_FILES, tutils.ROOTDIR, rulefile)
        os.remove(OUT_FILENAME)

    def test_print_ast(self):
        """Test that fsltool --ast runs without errors."""
        rulefile = tutils.get_fsltool_file('valid_complex')
        output = self._call_fsltool(['--ast', '-r', tutils.ROOTDIR, rulefile])
        self._check_enough_output(1000, output, 'fsltool --ast')

    def test_print_detailed(self):
        """Test that fsltool -d runs without errors."""
        rulefile = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-d', '-r', tutils.ROOTDIR, rulefile])
        self._check_enough_output(800, output, 'fsltool -d')

    def test_print_stats(self):
        """Test that fsltool -s runs without errors."""
        rulefile = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-s', '-r', tutils.ROOTDIR, rulefile])
        self._check_enough_output(800, output, 'fsltool -s')

    def test_cascade_nofiles(self):
        """Test that fsltool works with cascading inputs."""
        rulefile1 = tutils.get_fsltool_file('valid_allfiles')
        rulefile2 = tutils.get_fsltool_file('valid_nofiles')
        output = self._call_fsltool(['-r', tutils.ROOTDIR, rulefile1, rulefile2])
        filenames = output.splitlines()
        self.failIf(len(filenames) >= 1,
            'fsltool produced %s files, 0 expected' % len(filenames))

    def test_cascade_allfiles(self):
        """Test that fsltool works with cascading inputs."""
        rulefile1 = tutils.get_fsltool_file('valid_nofiles')
        rulefile2 = tutils.get_fsltool_file('valid_allfiles')
        output = self._call_fsltool(['-r', tutils.ROOTDIR, rulefile1, rulefile2])
        filenames = output.splitlines()
        self._check_filenames(filenames, test_interpreter.ALL_FILES, tutils.ROOTDIR, rulefile1+'+'+rulefile2)

def main(args):
    unittest.main()

if __name__=='__main__':
    import sys
    main(sys.argv[1:])
